<?php

namespace DBTech\Security\Admin\Controller;

use DBTech\Security\Helper;
use XF\Admin\Controller\AbstractController;
use XF\Mvc\ParameterBag;
use XF\Mvc\FormAction;

/**
 * Class Watcher
 * @package DBTech\Security\Admin\Controller
 */
class Watcher extends AbstractController
{
	/**
	 * @param $action
	 * @param ParameterBag $params
	 * @throws \XF\Mvc\Reply\Exception
	 */
	protected function preDispatchController($action, ParameterBag $params)
	{
		$this->assertAdminPermission('dbtechSecurity');
	}

	/**
	 * @return \XF\Mvc\Reply\AbstractReply
	 * @throws \Exception
	 */
	public function actionIndex(): \XF\Mvc\Reply\AbstractReply
	{
		$watchers = Helper::repository(\DBTech\Security\Repository\Watcher::class)
			->findWatchersForList()
			->fetch()
		;

		$viewParams = [
			'watchers' => $watchers
		];
		return $this->view(
			\DBTech\Security\Admin\View\Watcher\Listing::class,
			'dbtech_security_watcher_list',
			$viewParams
		);
	}

	/**
	 * @param \DBTech\Security\Entity\Watcher $watcher
	 * @return \XF\Mvc\Reply\AbstractReply
	 */
	protected function watcherAddEdit(\DBTech\Security\Entity\Watcher $watcher): \XF\Mvc\Reply\AbstractReply
	{
		$viewParams = [
			'watcher' => $watcher
		];
		return $this->view(
			\DBTech\Security\Admin\View\Watcher\Edit::class,
			'dbtech_security_watcher_edit',
			$viewParams
		);
	}

	/**
	 * @param ParameterBag $params
	 * @return \XF\Mvc\Reply\AbstractReply
	 * @throws \XF\Mvc\Reply\Exception
	 */
	public function actionEdit(ParameterBag $params): \XF\Mvc\Reply\AbstractReply
	{
		/** @var \DBTech\Security\Entity\Watcher $watcher */
		$watcher = $this->assertWatcherExists($params->watcher_id);
		return $this->watcherAddEdit($watcher);
	}

	/**
	 * @return \XF\Mvc\Reply\AbstractReply
	 * @throws \Exception
	 */
	public function actionAdd(): \XF\Mvc\Reply\AbstractReply
	{
		$watcherType = $this->filter('watcher_type', 'str');

		if ($watcherType)
		{
			$handler = Helper::repository(\DBTech\Security\Repository\Watcher::class)
				->getHandler($watcherType)
			;
			if ($handler)
			{
				$watcher = Helper::createEntity(\DBTech\Security\Entity\Watcher::class);
				$watcher->watcher_type = $watcherType;

				return $this->watcherAddEdit($watcher);
			}
		}

		$viewParams = [
			'watcherTypes' => Helper::repository(\DBTech\Security\Repository\Watcher::class)
				->getWatcherTitlePairs(true)
		];
		return $this->view(
			\DBTech\Security\Admin\View\Watcher\AddChooser::class,
			'dbtech_security_watcher_add_chooser',
			$viewParams
		);
	}

	/**
	 * @param \DBTech\Security\Entity\Watcher $watcher
	 *
	 * @return FormAction
	 * @throws \Exception
	 */
	protected function watcherSaveProcess(\DBTech\Security\Entity\Watcher $watcher): FormAction
	{
		$form = $this->formAction();

		$input = $this->filter([
			'active' => 'bool',
			'watcher_type' => 'str',
			'priority' => 'uint',
			'actions' => [
				'closeForum' => 'bool',
				'emailWebmaster' => 'bool',
				'banIp' => 'bool',
				'banUser' => 'bool',
				'emailUser' => 'bool',
				'lockChange' => 'bool',
				'lockReset' => 'bool',
				'lockUser' => 'bool',
				'adminLockUser' => 'bool'
			]
		]);

		$watcherType = Helper::repository(\DBTech\Security\Repository\Watcher::class)
			->getHandler($input['watcher_type'])
		;
		$input['rule_data'] = $watcherType->filterOptions($this->filter('rules', 'array'));

		$form->basicEntitySave($watcher, $input);

		return $form;
	}

	/**
	 * @param ParameterBag $params
	 *
	 * @return \XF\Mvc\Reply\AbstractReply
	 * @throws \XF\Mvc\Reply\Exception
	 * @throws \XF\PrintableException
	 * @throws \Exception
	 */
	public function actionSave(ParameterBag $params): \XF\Mvc\Reply\AbstractReply
	{
		$this->assertPostOnly();

		if ($params->watcher_id)
		{
			/** @var \DBTech\Security\Entity\Watcher $watcher */
			$watcher = $this->assertWatcherExists($params->watcher_id);
		}
		else
		{
			$watcher = Helper::createEntity(\DBTech\Security\Entity\Watcher::class);
		}

		$this->watcherSaveProcess($watcher)->run();

		return $this->redirect($this->buildLink('dbtech-security/watchers') . $this->buildLinkHash($watcher->watcher_id));
	}

	/**
	 * @param ParameterBag $params
	 *
	 * @return \XF\Mvc\Reply\AbstractReply
	 * @throws \XF\Mvc\Reply\Exception
	 */
	public function actionDelete(ParameterBag $params): \XF\Mvc\Reply\AbstractReply
	{
		$watcher = $this->assertWatcherExists($params->watcher_id);

		/** @var \XF\ControllerPlugin\Delete $plugin */
		$plugin = $this->plugin('XF:Delete');
		return $plugin->actionDelete(
			$watcher,
			$this->buildLink('dbtech-security/watchers/delete', $watcher),
			$this->buildLink('dbtech-security/watchers/edit', $watcher),
			$this->buildLink('dbtech-security/watchers'),
			$watcher->title
		);
	}

	/**
	 * @return \XF\Mvc\Reply\AbstractReply
	 */
	public function actionToggle(): \XF\Mvc\Reply\AbstractReply
	{
		/** @var \XF\ControllerPlugin\Toggle $plugin */
		$plugin = $this->plugin('XF:Toggle');
		return $plugin->actionToggle('DBTech\Security:Watcher');
	}

	/**
	 * @param int|null $id
	 * @param array|string|null $with
	 * @param null|string $phraseKey
	 *
	 * @return \DBTech\Security\Entity\Watcher
	 * @throws \XF\Mvc\Reply\Exception
	 */
	protected function assertWatcherExists(?int $id, $with = null, ?string $phraseKey = null): \DBTech\Security\Entity\Watcher
	{
		return $this->assertRecordExists('DBTech\Security:Watcher', $id, $with, $phraseKey);
	}
}