<?php

namespace DBTech\Security\Entity;

use DBTech\Security\Helper;
use XF\Mvc\Entity\Entity;
use XF\Mvc\Entity\Structure;
use DBTech\Security\Watcher\AbstractHandler;

/**
 * COLUMNS
 * @property int|null $watcher_id
 * @property string $watcher_type
 * @property bool $active
 * @property int $priority
 * @property array $rule_data
 * @property array $actions
 * @property array $extra_data
 *
 * GETTERS
 * @property \XF\Phrase $title
 * @property \XF\Phrase $description
 * @property string|\XF\Phrase $hint
 *
 * RELATIONS
 * @property \XF\Mvc\Entity\AbstractCollection|\DBTech\Security\Entity\WatcherLog[] $LogEntries
 */
class Watcher extends Entity
{
	/**
	 * @return bool
	 */
	public function canView(): bool
	{
		return $this->isActive();
	}

	/**
	 * @return bool
	 */
	public function isActive(): bool
	{
		return $this->active;
	}

	/**
	 * @return \XF\Phrase
	 * @throws \Exception
	 */
	public function getTitle(): \XF\Phrase
	{
		return $this->getWatcherHandler()->getTitle();
	}

	/**
	 * @return \XF\Phrase
	 * @throws \Exception
	 */
	public function getDescription(): \XF\Phrase
	{
		return $this->getWatcherHandler()->getDescription();
	}

	/**
	 * @return string|\XF\Phrase
	 * @throws \Exception
	 */
	public function getHint()
	{
		if (!empty($this->rule_data['field']))
		{
			return \XF::phrase('dbtech_security_watcher_field.' . $this->rule_data['field']);
		}

		return '';
	}

	/**
	 * @return \XF\Phrase
	 * @throws \Exception
	 */
	public function getParsedRule(): \XF\Phrase
	{
		return $this->getWatcherHandler()->getParsedRule($this);
	}

	/**
	 * @return array
	 * @throws \Exception
	 */
	public function getParsedActionSet(): array
	{
		$actions = [];

		if ($this->actions['closeForum'])
		{
			$actions[] = \XF::phrase('dbtech_security_actions.closeForum');
		}

		if ($this->actions['emailWebmaster'])
		{
			$actions[] = \XF::phrase('dbtech_security_actions.emailWebmaster');
		}

		if ($this->actions['banIp'])
		{
			$actions[] = \XF::phrase('dbtech_security_actions.banIp');
		}

		if ($this->actions['banUser'])
		{
			$actions[] = \XF::phrase('dbtech_security_actions.banUser');
		}

		if ($this->actions['emailUser'])
		{
			$actions[] = \XF::phrase('dbtech_security_actions.emailUser');
		}

		if ($this->actions['lockChange'])
		{
			$actions[] = \XF::phrase('dbtech_security_actions.lockChange');
		}

		if ($this->actions['lockReset'])
		{
			$actions[] = \XF::phrase('dbtech_security_actions.lockReset');
		}

		if ($this->actions['lockUser'])
		{
			$actions[] = \XF::phrase('dbtech_security_actions.lockUser');
		}

		if ($this->actions['adminLockUser'])
		{
			$actions[] = \XF::phrase('dbtech_security_actions.adminLockUser');
		}

		return $actions;
	}

	/**
	 * @param bool $throw
	 *
	 * @return AbstractHandler|null
	 * @throws \Exception
	 */
	public function getWatcherHandler(bool $throw = true): ?AbstractHandler
	{
		return Helper::repository(\DBTech\Security\Repository\Watcher::class)
			->getHandler($this->watcher_type, $throw)
		;
	}

	/**
	 * @return bool
	 */
	protected function _preSave(): bool
	{
		$exists = Helper::finder(\DBTech\Security\Finder\Watcher::class)
			->where('watcher_type', $this->watcher_type)
			->where('rule_data', $this->getValueSourceEncoded('rule_data'))
			->fetchOne();
		if ($exists && $exists !== $this)
		{
			$this->error(\XF::phrase('dbtech_security_specified_watcher_rule_set_already_exists'));
			return false;
		}

		return true;
	}

	/**
	 *
	 */
	protected function _postDelete()
	{
		$db = $this->db();
		$db->delete('xf_dbtech_security_watcher_log', 'watcher_id = ?', $this->watcher_id);
	}

	/**
	 * @param \XF\Mvc\Entity\Structure $structure
	 *
	 * @return \XF\Mvc\Entity\Structure
	 */
	public static function getStructure(Structure $structure): Structure
	{
		$structure->table = 'xf_dbtech_security_watcher';
		$structure->shortName = 'DBTech\Security:Watcher';
		$structure->primaryKey = 'watcher_id';
		$structure->columns = [
			'watcher_id' => ['type' => self::UINT, 'autoIncrement' => true, 'nullable' => true],
			'watcher_type' => ['type' => self::STR, 'required' => true],
			'active' => ['type' => self::BOOL, 'default' => true],
			'priority' => ['type' => self::UINT, 'default' => 1],
			'rule_data' => ['type' => self::JSON_ARRAY, 'default' => []],
			'actions' => ['type' => self::JSON_ARRAY, 'default' => []],
			'extra_data' => ['type' => self::JSON_ARRAY, 'default' => []],
		];
		$structure->getters = [
			'title' => true,
			'description' => true,
			'hint' => true,
			'explain' => true
		];
		$structure->behaviors = [
			'DBTech\Security:Cacheable' => []
		];
		$structure->relations = [
			'LogEntries' => [
				'entity' => 'DBTech\Security:WatcherLog',
				'type' => self::TO_MANY,
				'conditions' => 'watcher_id'
			]
		];

		return $structure;
	}
}