<?php

namespace DBTech\Security;

use DBTech\Security\Repository\Watcher as WatcherRepo;
use XF\Container;

class Listener
{
	/**
	 * The product ID (in the DBTech store)
	 * @var int
	 */
	protected static int $_productId = 345;

	/**
	 * @param \XF\Pub\App $app
	 */
	public static function appPubSetup(\XF\Pub\App $app): void
	{
		/*DBTECH_BRANDING_START*/
		// Make sure we fetch the branding array from the application
		$branding = $app->offsetExists('dbtech_branding') ? $app->dbtech_branding : [];

		// Add productid to the array
		$branding[] = self::$_productId;

		// Store the branding
		$app->dbtech_branding = $branding;
		/*DBTECH_BRANDING_END*/
	}

	/**
	 * @param \XF\App $app
	 *
	 * @throws \XF\Db\Exception
	 */
	public static function appSetup(\XF\App $app): void
	{
		$container = $app->container();

		$container['dbtechSecurity.watchers'] = $app->fromRegistry(
			'dbtSecurityWatchers',
			function (Container $c)
			{
				return $c['em']->getRepository('DBTech\Security:Watcher')->rebuildCache();
			},
			function (array $watchers): array
			{
				$em = \XF::em();

				$entities = [];
				foreach ($watchers as $watcherId => $watcher)
				{
					if (!isset($entities[$watcher['watcher_type']]))
					{
						$entities[$watcher['watcher_type']] = [];
					}

					$watcher['rule_data'] = json_encode($watcher['rule_data']);
					$watcher['actions'] = json_encode($watcher['actions']);
					$watcher['extra_data'] = json_encode($watcher['extra_data']);

					$entities[$watcher['watcher_type']][$watcherId] = Helper::instantiateEntity(\DBTech\Security\Entity\Watcher::class, $watcher);
				}

				foreach ($entities as $watcher => $forCollection)
				{
					$entities[$watcher] = $em->getBasicCollection($forCollection);
				}

				return $entities;
			}
		);
	}

	/**
	 * @param \XF\Pub\App $app
	 *
	 * @throws \XF\Db\Exception
	 * @throws \XF\Db\Exception
	 */
	public static function appPubStartEnd(\XF\Pub\App $app): void
	{
		$sessionRepo = Helper::repository(\DBTech\Security\Repository\Session::class);
		$sessionId = $sessionRepo->getSessionCookie();
		if (!$sessionId)
		{
			// Create a new session
			$sessionRepo->createSession(\XF::visitor());
		}
		else
		{
			if (!Helper::find(\DBTech\Security\Entity\Session::class, $sessionId))
			{
				$sessionRepo->deleteSessionCookie();

				$app->session()->logoutUser();

				$user = Helper::repository(\XF\Repository\User::class)->getGuestUser();
				\XF::setVisitor($user);
			}
		}
	}

	/**
	 * @param \XF\Template\Templater $templater
	 * @param string $type
	 * @param string $template
	 * @param string $name
	 * @param array $arguments
	 * @param array $globalVars
	 */
	public static function templaterMacroPreRender(
		\XF\Template\Templater $templater,
		string &$type,
		string &$template,
		string &$name,
		array &$arguments,
		array &$globalVars
	): void {
		if (!empty($arguments['group']) && $arguments['group']->group_id == 'dbtech_security')
		{
			// Override template name
			$template = 'dbtech_security_option_macros';
		}
	}

	/**
	 * @param \XF\Mvc\Controller $controller
	 * @param string $action
	 * @param \XF\Mvc\ParameterBag $params
	 *
	 * @throws \XF\Mvc\Reply\Exception
	 */
	public static function controllerPreDispatch(
		\XF\Mvc\Controller $controller,
		string $action,
		\XF\Mvc\ParameterBag $params
	): void {
		$app = \XF::app();
		$router = $app->router('public');

		/** @var \DBTech\Security\XF\Entity\User $visitor */
		$visitor = \XF::visitor();

		if ($visitor->user_id)
		{
			/*
			$clientIps = Ip::getClientIps();
			foreach (array_unique($clientIps) as $clientIp)
			{
				// Log IP
				$app->db()->insert('xf_dbtech_security_ip_log', [
					'user_id' => $visitor->user_id,
					'ipaddress' => $clientIp,
					'first_visit' => \XF::$time,
					'last_visit' => \XF::$time
				], false, 'last_visit = UNIX_TIMESTAMP()');
			}
			*/

			$options = \XF::options();
			$request = $app->request();
			$requestUri = $request->getFullRequestUri();

			if ($visitor->hasDbtechSecurityRequiredPasswordChange()
				&& !Util\Redirect::canBypassRedirect(
					$options->dbtechSecurityPasswordChangeForceWhitelist,
					$router->buildLink('full:account/security'),
					$requestUri
				)
			) {
				// We need a password change
				throw $controller->exception($controller->redirect($router->buildLink('account/security', null, [
					'_xfRedirect' => $request->getFullRequestUri()
				]), \XF::phrase('dbtech_security_need_password_change')));
			}

			if (
				(
					$visitor->Option->dbtech_security_is_user_locked
					|| $visitor->Option->dbtech_security_is_admin_locked
				)
				&& !Util\Redirect::canBypassRedirect(
					$options->dbtechSecurityAccountLockForceWhitelist,
					$router->buildLink('full:dbtech-security/account-lock'),
					$requestUri
				)
			) {
				// We need a password change
				throw $controller->exception($controller->redirect($router->buildLink('dbtech-security/account-lock', null, [
					'_xfRedirect' => $request->getFullRequestUri()
				]), \XF::phrase('dbtech_security_account_locked')));
			}
		}

		
	}

	/**
	 * @param \XF\Mvc\Dispatcher $dispatcher
	 * @param string|null $content
	 * @param \XF\Mvc\Reply\AbstractReply $reply
	 * @param \XF\Mvc\Renderer\AbstractRenderer $renderer
	 * @param \XF\Http\Response $response
	 *
	 * @throws \Exception
	 */
	public static function dispatcherPostRender(
		\XF\Mvc\Dispatcher $dispatcher,
		?string &$content,
		\XF\Mvc\Reply\AbstractReply $reply,
		\XF\Mvc\Renderer\AbstractRenderer $renderer,
		\XF\Http\Response $response
	): void {
		$app = \XF::app();
		$options = $app->options();

		/** @var \DBTech\Security\XF\Entity\User $visitor */
		$visitor = \XF::visitor();

		if ($app['app.classType'] != 'Pub'
			|| $response->contentType() != 'text/html'
			|| $dispatcher->getRequest()->isPost()
			|| in_array($reply->getControllerClass(), [
				'XF:Attachment',
				'XF:Login',
				'XF:Logout'
			])
		) {
			// We're not doing this
			return;
		}

		if ($options->dbtech_security_badbehavior_enabled && !WatcherRepo::noBadBehavior())
		{
			$bb2_run = true;

			if ($visitor->user_id)
			{
				if ($options->dbtech_security_badbehavior_bypass_user['enabled']
					&& (
						$options->dbtech_security_badbehavior_bypass_user['days'] < 1
						|| ((\XF::$time - $visitor->register_date / 86400) > $options->dbtech_security_badbehavior_bypass_user['days'])
					)
				) {
					$bb2_run = false;
				}
			}

			if ($bb2_run === true)
			{
				/** @noinspection PhpIncludeInspection */
				require_once(\XF::getAddOnDirectory() . '/DBTech/Security/3rdParty/functions_badbehavior.php');

				bb2_start(bb2_read_settings());

				if (!empty($GLOBALS['bb2_javascript']))
				{
					$content = str_replace('</head>', $GLOBALS['bb2_javascript'] . '</head>', $content);
				}
			}
		}

		$watcherRepo = Helper::repository(\DBTech\Security\Repository\Watcher::class);
		$handler = $watcherRepo->getHandler('configtamper', false);

		if ($handler)
		{
			$handler->trigger();
		}
	}

	/**
	 * @param \XF\Service\User\DeleteCleanUp $deleteService
	 * @param array $deletes
	 */
	public static function userDeleteCleanInit(\XF\Service\User\DeleteCleanUp $deleteService, array &$deletes): void
	{
		$deletes = array_merge($deletes, [
			'xf_dbtech_security_account_lock' => 'user_id = ?',
			'xf_dbtech_security_admin_strike' => 'user_id = ?',
			'xf_dbtech_security_change_log' => 'user_id = ?',
			'xf_dbtech_security_compromised_log' => 'user_id = ?',
			'xf_dbtech_security_fingerprint_log' => 'user_id = ?',
			'xf_dbtech_security_ip_log' => 'user_id = ?',
			'xf_dbtech_security_ip_verify' => 'user_id = ?',
			'xf_dbtech_security_recovery_log' => 'user_id = ?',
			'xf_dbtech_security_session' => 'user_id = ?',
			'xf_dbtech_security_watcher_log' => 'user_id = ?',
		]);
	}
}