<?php

namespace DBTech\Security\Repository;

use DBTech\Security\Helper;
use XF\Mvc\Entity\Finder;
use XF\Mvc\Entity\Repository;
use XF\Util\File;

class Banning extends Repository
{
	/**
	 * @return Finder
	 */
	public function findTorExitNodes(): Finder
	{
		return Helper::finder('XF:IpMatch')
			->where('match_type', 'dbtech_security_tor')
			->setDefaultOrder('start_range', 'asc');
	}

	/**
	 * @param string $ip
	 *
	 * @return bool
	 * @throws \XF\PrintableException
	 */
	public function addTorExitNode(string $ip): bool
	{
		[$niceIp, $firstByte, $startRange, $endRange] = $this->getIpRecord($ip);

		$ipBan = Helper::createEntity(\XF\Entity\IpMatch::class);
		$ipBan->ip = $niceIp;
		$ipBan->match_type = 'dbtech_security_tor';
		$ipBan->first_byte = $firstByte;
		$ipBan->start_range = $startRange;
		$ipBan->end_range = $endRange;
		$ipBan->reason = '[DragonByte Security] TOR Exit Node';
		$ipBan->create_user_id = 0;

		return $ipBan->save();
	}

	/**
	 *
	 */
	public function removeBannedTorIps()
	{
		$this->db()->delete('xf_ip_match', 'match_type = \'banned\' AND dbtech_security_comment = \'tor\'');

		\XF::runOnce('bannedIpCache', function ()
		{
			Helper::repository(\XF\Repository\Banning::class)
				->rebuildBannedIpCache()
			;
		});
	}

	/**
	 *
	 */
	public function rebuildBannedTorIps()
	{
		$db = $this->db();
		$db->beginTransaction();

		$this->removeBannedTorIps();

		$torNodes = $this->findTorExitNodes()
			->fetch();

		$insert = [];
		foreach ($torNodes AS $torNode)
		{
			$entry = $torNode->toArray();
			$entry['match_type'] = 'banned';
			$entry['dbtech_security_comment'] = 'tor';

			$insert[] = $entry;
		}

		if ($insert)
		{
			$this->db()->insertBulk('xf_ip_match', $insert, false, false, 'IGNORE');
		}

		\XF::runOnce('bannedIpCache', function ()
		{
			Helper::repository(\XF\Repository\Banning::class)
				->rebuildBannedIpCache()
			;
		});

		$db->commit();
	}

	/**
	 * @return Finder
	 */
	public function findCountryIps():Finder
	{
		return Helper::finder('XF:IpMatch')
			->where('match_type', 'dbtech_security_country')
			->setDefaultOrder('start_range', 'asc');
	}


	/**
	 * @param string $countryCode
	 * @param string $ip
	 *
	 * @return bool
	 * @throws \XF\PrintableException
	 */
	public function banCountryIp(string $countryCode, string $ip): bool
	{
		[$niceIp, $firstByte, $startRange, $endRange] = $this->getIpRecord($ip);

		$ipBan = Helper::createEntity(\XF\Entity\IpMatch::class);
		$ipBan->ip = $niceIp;
		$ipBan->match_type = 'banned';
		$ipBan->dbtech_security_comment = $countryCode;
		$ipBan->first_byte = $firstByte;
		$ipBan->start_range = $startRange;
		$ipBan->end_range = $endRange;
		$ipBan->reason = '[DragonByte Security] Blocked country';
		$ipBan->create_user_id = 0;

		return $ipBan->save();
	}

	/**
	 * @param string $countryCode
	 */
	public function removeBannedCountryIps(string $countryCode)
	{
		$this->db()->delete(
			'xf_ip_match',
			"match_type = 'banned' AND dbtech_security_comment = ?",
			$countryCode
		);

		\XF::runOnce('bannedIpCache', function ()
		{
			Helper::repository(\XF\Repository\Banning::class)
				->rebuildBannedIpCache()
			;
		});
	}

	/**
	 * @param string $countryCode
	 *
	 * @throws \XF\PrintableException
	 */
	public function updateAndBanIpsForCountry(string $countryCode)
	{
		/** @var Banning $banningRepo */
		$banningRepo = Helper::repository(Banning::class);

		$tempFile = File::getTempFile();

		/** @noinspection HttpUrlsUsage */
		$url = 'http://www.ipdeny.com/ipblocks/data/aggregated/' . strtolower($countryCode) . '-aggregated.zone';

		$response = $this->app()
			->http()
			->reader()
			->getUntrusted($url, [], $tempFile)
		;
		if (!$response || $response->getStatusCode() != 200)
		{
			return;
		}

		// Remove existing countries
		$banningRepo->removeBannedCountryIps($countryCode);

		$ips = preg_split('/\r?\n/', file_get_contents($tempFile), -1, PREG_SPLIT_NO_EMPTY);
		foreach ($ips as $ip)
		{
			// Add new country IP
			$banningRepo->banCountryIp($countryCode, $ip);
		}
	}

	/**
	 * @param string $ip
	 *
	 * @return array
	 * @throws \XF\PrintableException
	 */
	protected function getIpRecord(string $ip): array
	{
		$results = \XF\Util\Ip::parseIpRangeString($ip);
		if (!$results)
		{
			throw new \XF\PrintableException(\XF::phrase('please_enter_valid_ip_or_ip_range'));
		}

		return [
			$results['printable'],
			$results['binary'][0],
			$results['startRange'],
			$results['endRange']
		];
	}
}