<?php

namespace DBTech\Security\Repository;

use DBTech\Security\Helper;
use XF\Mvc\Entity\Repository;

/**
 * Class Country
 * @package DBTech\Security\Repository
 */
class Country extends Repository
{
	/**
	 * @return \DBTech\Security\Finder\Country
	 */
	public function findCountriesForList(): \DBTech\Security\Finder\Country
	{
		return Helper::finder(\DBTech\Security\Finder\Country::class)
			->orderName();
	}

	/**
	 * @param string $countryCode
	 *
	 * @return \DBTech\Security\Finder\Country
	 */
	public function findCountryByCode(string $countryCode): \DBTech\Security\Finder\Country
	{
		return Helper::finder(\DBTech\Security\Finder\Country::class)
			->where('country_code', $countryCode);
	}

	/**
	 * @param bool $includeEmpty
	 * @param string|null $type
	 *
	 * @return array
	 */
	public function getCountryOptionsData(bool $includeEmpty = true, ?string $type = null): array
	{
		$choices = [];
		if ($includeEmpty)
		{
			$choices = [
				'' => ['value' => '', 'label' => \XF::phrase('(none)')]
			];
			if ($type !== null)
			{
				$choices['']['_type'] = $type;
			}
		}

		$countryList = $this->findCountriesForList();

		/** @var \DBTech\Security\Entity\Country $entry */
		foreach ($countryList AS $entry)
		{
			$choices[$entry->country_code] = [
				'value' => $entry->country_code,
				'label' => $entry->name
			];
			if ($type !== null)
			{
				$choices[$entry->country_code]['_type'] = $type;
			}
		}

		return $choices;
	}

	/**
	 * @param bool $includeEmpty
	 * @param bool $includeAll
	 *
	 * @return array
	 */
	public function getCountrySelectData(bool $includeEmpty = true, bool $includeAll = false): array
	{
		$choices = [];
		if ($includeEmpty)
		{
			$choices = [
				'' => \XF::phrase('(none)')
			];
		}
		if ($includeAll)
		{
			$choices = [
				'-1' => \XF::phrase('(all)')
			];
		}

		$countryList = $this->findCountriesForList();

		/** @var \DBTech\Security\Entity\Country $entry */
		foreach ($countryList AS $entry)
		{
			$choices[$entry->country_code] = $entry->name;
		}

		return $choices;
	}

	/**
	 * @return bool
	 * @throws \Exception
	 * @throws \XF\PrintableException
	 */
	public function updateCountryList(): bool
	{
		$reader = $this->app()
			->http()
			->reader()
		;

		$countryList = null;

		try
		{
			/** @noinspection HttpUrlsUsage */
			$response = $reader->getUntrusted(
				'https://raw.githubusercontent.com/mledoze/countries/master/dist/countries.json'
			);
			if ($response)
			{
				$jsonText = $response->getBody()->getContents();

				$response->getBody()->close();

				if ($response->getStatusCode() == 200)
				{
					try
					{
						$countryList = \GuzzleHttp\json_decode($jsonText, true);
					}
					catch (\InvalidArgumentException $e)
					{
						\XF::logException($e, false, 'Security error:');
						return false;
					}
				}
				else
				{
					\XF::logError(\XF::phraseDeferred('received_unexpected_response_code_x_message_y', [
						'code' => $response->getStatusCode(),
						'message' => $response->getReasonPhrase()
					]));
					return false;
				}
			}
		}
		catch (\GuzzleHttp\Exception\RequestException $e)
		{
			\XF::logException($e, false, 'Security error:');
			return false;
		}

		if (!is_array($countryList))
		{
			return false;
		}

		/** @var \DBTech\Security\Entity\Country[]|\XF\Mvc\Entity\ArrayCollection $existingCountries */
		$existingCountries = $this->findCountriesForList()->fetch();

		$currentCountries = [];
		foreach ($countryList as $country)
		{
			$nativeName = \reset($country['name']['native']);
			if ($nativeName === false)
			{
				$nativeName = $country['name'];
			}

			/** @var \DBTech\Security\Entity\Country $newCountry */
			if (!isset($existingCountries[$country['cca2']]))
			{
				$newCountry = Helper::createEntity(\DBTech\Security\Entity\Country::class);
				$newCountry->bulkSet([
					'country_code' => $country['cca2'],
					'name' => $country['name']['common'],
					'native_name' => $nativeName['common'],
					'iso_code' => $country['cca3'],
				]);
				$newCountry->save();
			}
			else
			{
				$newCountry = $existingCountries[$country['cca2']];

				if ($newCountry->name != $country['name']['common'])
				{
					$newCountry->name = $country['name']['common'];
				}

				if ($newCountry->native_name != $nativeName['common'])
				{
					$newCountry->native_name = $nativeName['common'];
				}

				$newCountry->saveIfChanged();
			}
			$currentCountries[$country['cca2']] = $newCountry;
		}

		$missingCountries = array_diff_key($existingCountries->toArray(), $currentCountries);

		/** @var \DBTech\Security\Entity\Country $country */
		foreach ($missingCountries as $country)
		{
			$country->delete();
		}

		return true;
	}
}