<?php

namespace DBTech\Security\Repository;

use DBTech\Security\Helper;
use XF\Mvc\Entity\Repository;

/**
 * Class Session
 * @package DBTech\Security\Repository
 */
class Session extends Repository
{
	/**
	 * @param \XF\Entity\User|null $user
	 *
	 * @return \XF\Mvc\Entity\Finder
	 */
	public function getSessionsForUser(?\XF\Entity\User $user = null): \XF\Mvc\Entity\Finder
	{
		$user = $user ?: \XF::visitor();

		return Helper::finder(\DBTech\Security\Finder\Session::class)
			->where('user_id', $user->user_id)
			->setDefaultOrder('start_date')
		;
	}

	/**
	 * Invalidate sessions older than X days (default: 30)
	 *
	 * @param int $days
	 */
	public function prune(int $days = 30)
	{
		$this->db()->delete(
			'xf_dbtech_security_session',
			'last_activity_date < ?',
			[\XF::$time - ($days * 86400)]
		)
		;
	}

	/**
	 * Get a session by session ID
	 *
	 * @param string $sessionId
	 *
	 * @return array|bool
	 */
	public function getSessionBySessionId(string $sessionId)
	{
		return $this->db()->fetchRow('
			SELECT *
			FROM xf_dbtech_security_session
			WHERE session_id = ?
		', [$sessionId]);
	}

	/**
	 * Creates a new session for the current user
	 *
	 * @param \XF\Entity\User $user
	 * @param bool $setCookie
	 *
	 * @return \DBTech\Security\Entity\Session
	 * @throws \XF\Db\Exception
	 */
	public function createSession(\XF\Entity\User $user, bool $setCookie = true): \DBTech\Security\Entity\Session
	{
		$sessionData = [
			'session_id' => md5(\XF::generateRandomString(16, true)),
			'user_id' => $user->user_id,
			'start_date' => \XF::$time,
			'last_activity_date' => \XF::$time,
			'user_agent' => $this->app()->request()->getUserAgent() ?: 'N/A',
		];

		$query = "-- XFDB=noForceAllWrite
			INSERT IGNORE INTO xf_dbtech_security_session (
				session_id,
				user_id,
				start_date,
				last_activity_date,
				user_agent
			) VALUES (?, ?, ?, ?, ?)
		";

		$this->db()->query($query, $sessionData);

		/** @var \DBTech\Security\Entity\Session $session */
		$session = Helper::instantiateEntity(\DBTech\Security\Entity\Session::class, $sessionData);

		if ($setCookie)
		{
			$this->setSessionCookie($session->session_id);
		}

		return $session;
	}

	/**
	 * Update session last activity date
	 *
	 * @param \DBTech\Security\Entity\Session $session
	 * @param \XF\Entity\User|null $user
	 *
	 * @throws \XF\Db\Exception
	 */
	public function updateSessionLastActive(\DBTech\Security\Entity\Session $session, ?\XF\Entity\User $user = null)
	{
		$ua = $this->app()->request()->getUserAgent() ?: 'N/A';

		if ($user)
		{
			$this->db()->query('
				-- XFDB=noForceAllWrite
				INSERT INTO xf_dbtech_security_session
					(session_id, user_id, last_activity_date, user_agent)
				VALUES
					(?, ?, ?, ?)
				ON DUPLICATE KEY UPDATE
					user_id = VALUES(user_id),
					last_activity_date = VALUES(last_activity_date),
					user_agent = VALUES(user_agent)
			', [$session->session_id, $user->user_id, \XF::$time, $ua]);
		}
		else
		{
			$this->db()->query('
				-- XFDB=noForceAllWrite
				INSERT INTO xf_dbtech_security_session
					(session_id, last_activity_date, user_agent)
				VALUES
					(?, ?, ?)
				ON DUPLICATE KEY UPDATE
					last_activity_date = VALUES(last_activity_date),
					user_agent = VALUES(user_agent)
			', [$session->session_id, \XF::$time, $ua]);
		}

		$this->setSessionCookie($session->session_id);
	}

	/**
	 * Get a session cookie
	 *
	 * @return array|bool|string
	 */
	public function getSessionCookie()
	{
		return $this->app()->request()->getCookie('dbtechSecuritySession');
	}

	/**
	 * Set session cookie
	 *
	 * @param string $sessionId
	 * @param int $lifetimeDays
	 */
	public function setSessionCookie(string $sessionId, int $lifetimeDays = 30): void
	{
		$this->app()->response()->setCookie('dbtechSecuritySession', $sessionId, $lifetimeDays * 86400);
	}

	/**
	 * Delete session cookie
	 */
	public function deleteSessionCookie(): void
	{
		$this->app()->response()->setCookie('dbtechSecuritySession', false);
	}
}