<?php

namespace DBTech\Security\Tfa;

use DBTech\Security\Helper;
use XF\Entity\TfaProvider;
use XF\PrintableException;
use XF\Tfa\AbstractProvider;
use Webauthn\AuthenticatorSelectionCriteria;
use Webauthn\PublicKeyCredentialCreationOptions;
use Webauthn\PublicKeyCredentialRpEntity;
use Webauthn\PublicKeyCredentialSource;
use Webauthn\PublicKeyCredentialUserEntity;
use Webauthn\PublicKeyCredentialRequestOptions;
use Webauthn\Server;
use Webauthn\MetadataService\SimpleMetadataStatementRepository;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\Cache\Adapter\FilesystemAdapter;
use Nyholm\Psr7\Factory\Psr17Factory;
use Symfony\Bridge\PsrHttpMessage\Factory\PsrHttpFactory;
use DBTech\Security\WebAuthn\PublicKeyCredentialSourceRepositoryDb as PublicKeyCredentialSourceRepository;

/**
 * Class WebAuthn
 *
 * @package DBTech\Security\Tfa
 */
class WebAuthn extends AbstractProvider
{
	protected bool $enableMds = false;


	/**
	 * @return bool
	 */
	public function canEnable(): bool
	{
		return (PHP_VERSION_ID >= 70200);
	}

	/**
	 * @return bool
	 */
	public function isUsable(): bool
	{
		return (PHP_VERSION_ID >= 70200 && \XF::app()->request()->isSecure());
	}

	/**
	 * @param TfaProvider $provider
	 *
	 * @return string
	 */
	public function renderOptions(TfaProvider $provider): string
	{
		$params = [
			'provider' => $provider
		];
		return \XF::app()->templater()->renderTemplate('admin:dbtech_security_two_step_config_webauthn', $params);
	}

	/**
	 * @param \XF\Entity\User $user
	 * @param array $config
	 *
	 * @return array
	 */
	public function generateInitialData(\XF\Entity\User $user, array $config = []): array
	{
		$provider = $this->getProvider();
		$userConfig = $provider->getUserProviderConfig();

		$credentials = !empty($userConfig['credentialStorage']) ? $userConfig['credentialStorage'] : [];
		$repo = new PublicKeyCredentialSourceRepository($credentials);

		// Credentials server
		$server = $this->getCredentialsServer($repo);

		// User Entity
		$userEntity = $this->getUserEntity($user);

		$registeredPublicKeyCredentialSources = $repo->findAllForUserEntity($userEntity);
		$registeredPublicKeyCredentialDescriptors = array_map(static function (PublicKeyCredentialSource $item): \Webauthn\PublicKeyCredentialDescriptor
		{
			return $item->getPublicKeyCredentialDescriptor();
		}, $registeredPublicKeyCredentialSources);

		$publicKeyCredentialCreationOptions = $server->generatePublicKeyCredentialCreationOptions(
			$userEntity,
			$this->enableMds
				? PublicKeyCredentialCreationOptions::ATTESTATION_CONVEYANCE_PREFERENCE_DIRECT
				: PublicKeyCredentialCreationOptions::ATTESTATION_CONVEYANCE_PREFERENCE_NONE,
			$registeredPublicKeyCredentialDescriptors,
			new AuthenticatorSelectionCriteria(
				null,
				!empty($provider->options['key_pin']),
				!empty($provider->options['key_pin'])
					? AuthenticatorSelectionCriteria::USER_VERIFICATION_REQUIREMENT_PREFERRED
					: AuthenticatorSelectionCriteria::USER_VERIFICATION_REQUIREMENT_DISCOURAGED
			)
		);

		$encodedOptions = \GuzzleHttp\json_encode($publicKeyCredentialCreationOptions, JSON_UNESCAPED_SLASHES | JSON_UNESCAPED_UNICODE);

		return [
			'encodedOptions' => $encodedOptions
		];
	}

	/**
	 * @param $context
	 * @param \XF\Entity\User $user
	 * @param array $config
	 * @param \XF\Http\Request $request
	 *
	 * @return array
	 */
	public function trigger($context, \XF\Entity\User $user, array &$config, \XF\Http\Request $request): array
	{
		return [];
	}

	/**
	 * @param $context
	 * @param \XF\Entity\User $user
	 * @param array $config
	 * @param array $triggerData
	 *
	 * @return string
	 */
	public function render($context, \XF\Entity\User $user, array $config, array $triggerData): string
	{
		$encodedOptions = !empty($config['encodedOptions']) ? $config['encodedOptions'] : [];
		if ($context == 'login')
		{
			$sessionKey = 'tfaData_dbtech_security_authn';
			$session = \XF::session();

			$providerData = $session->get($sessionKey);
			if (!$providerData)
			{
				$provider = $this->getProvider();
				$credentials = !empty($config['credentialStorage']) ? $config['credentialStorage'] : [];
				$repo = new PublicKeyCredentialSourceRepository($credentials);

				// Credentials server
				$server = $this->getCredentialsServer($repo);

				// User Entity
				$userEntity = $this->getUserEntity($user);

				$registeredPublicKeyCredentialSources = $repo->findAllForUserEntity($userEntity);
				$registeredPublicKeyCredentialDescriptors = array_map(static function (PublicKeyCredentialSource $item): \Webauthn\PublicKeyCredentialDescriptor
				{
					return $item->getPublicKeyCredentialDescriptor();
				}, $registeredPublicKeyCredentialSources);

				// Public Key Credential Request Options
				$publicKeyCredentialRequestOptions = $server->generatePublicKeyCredentialRequestOptions(
					!empty($provider->options['key_pin'])
						? PublicKeyCredentialRequestOptions::USER_VERIFICATION_REQUIREMENT_PREFERRED
						: PublicKeyCredentialRequestOptions::USER_VERIFICATION_REQUIREMENT_DISCOURAGED,
					$registeredPublicKeyCredentialDescriptors
				);
				$encodedOptions = \GuzzleHttp\json_encode($publicKeyCredentialRequestOptions, JSON_UNESCAPED_SLASHES | JSON_UNESCAPED_UNICODE);

				$session->set($sessionKey, ['encodedOptions' => $encodedOptions]);
			}
			else
			{
				$encodedOptions = $providerData['encodedOptions'];
			}
		}

		$params = [
			'encodedOptions' => $encodedOptions,
			'config' => $config,
			'context' => $context
		];
		return \XF::app()->templater()->renderTemplate('public:dbtech_security_two_step_webauthn', $params);
	}

	/**
	 * @param $context
	 * @param \XF\Entity\User $user
	 * @param array $config
	 * @param \XF\Http\Request $request
	 *
	 * @return bool
	 * @throws PrintableException
	 */
	public function verify($context, \XF\Entity\User $user, array &$config, \XF\Http\Request $request): bool
	{
		if ($context == 'setup')
		{
			/** @var PublicKeyCredentialCreationOptions $publicKeyCredentialCreationOptions */
			$publicKeyCredentialCreationOptions = PublicKeyCredentialCreationOptions::createFromString($config['encodedOptions']);

			// Retrieve de data sent by the device
			$data = base64_decode($request->filter('publicKeyCredential', 'str'));

			$credentials = !empty($config['credentialStorage']) ? $config['credentialStorage'] : [];
			$repo = new PublicKeyCredentialSourceRepository($credentials);

			// Credentials server
			$server = $this->getCredentialsServer($repo);

			try
			{
				$symfonyRequest = Request::createFromGlobals();

				$psr17Factory = new Psr17Factory();
				$psrHttpFactory = new PsrHttpFactory($psr17Factory, $psr17Factory, $psr17Factory, $psr17Factory);
				$psr7Request = $psrHttpFactory->createRequest($symfonyRequest);

				// Check the response against the request
				$publicKeyCredentialSource = $server->loadAndCheckAttestationResponse($data, $publicKeyCredentialCreationOptions, $psr7Request);

				// Save to the credentialStorage of the `xf_user_tfa` record
				$config['credentialStorage'][base64_encode($publicKeyCredentialSource->getPublicKeyCredentialId())] = [
					'name' => $request->filter('nickname', 'str') ?: 'My security key',
					'register_date' => \XF::$time,
					'device' => $publicKeyCredentialSource,
				];

				unset($config['encodedOptions']);
			}
			catch (\Throwable $e)
			{
				\XF::logException($e);
				throw new PrintableException($e->getMessage());
			}
		}
		elseif ($context == 'login')
		{
			$sessionKey = 'tfaData_dbtech_security_authn';
			$session = \XF::session();
			$sessionData = $session->get($sessionKey);

			/** @var PublicKeyCredentialRequestOptions $publicKeyCredentialRequestOptions */
			$publicKeyCredentialRequestOptions =  PublicKeyCredentialRequestOptions::createFromString($sessionData['encodedOptions']);

			// Retrieve de data sent by the device
			$data = base64_decode($request->filter('publicKeyCredential', 'str'));

			$credentials = !empty($config['credentialStorage']) ? $config['credentialStorage'] : [];
			$repo = new PublicKeyCredentialSourceRepository($credentials);

			// Credentials server
			$server = $this->getCredentialsServer($repo);

			// User Entity
			$userEntity = $this->getUserEntity($user);

			try
			{
				$symfonyRequest = Request::createFromGlobals();

				$psr17Factory = new Psr17Factory();
				$psrHttpFactory = new PsrHttpFactory($psr17Factory, $psr17Factory, $psr17Factory, $psr17Factory);
				$psr7Request = $psrHttpFactory->createRequest($symfonyRequest);

				// Check the response against the request
				$server->loadAndCheckAssertionResponse(
					$data,
					$publicKeyCredentialRequestOptions,
					$userEntity,
					$psr7Request
				);

				$session->remove($sessionKey);
			}
			catch (\Throwable $e)
			{
				\XF::logException($e);
				throw new PrintableException($e->getMessage());
			}
		}

		return true;
	}

	/**
	 * @return bool
	 */
	public function canManage(): bool
	{
		return true;
	}

	/**
	 * @param \XF\Mvc\Controller $controller
	 * @param TfaProvider $provider
	 * @param \XF\Entity\User $user
	 * @param array $config
	 *
	 * @return \XF\Mvc\Reply\AbstractReply|null
	 * @throws PrintableException
	 */
	public function handleManage(
		\XF\Mvc\Controller $controller,
		TfaProvider $provider,
		\XF\Entity\User $user,
		array $config
	): ?\XF\Mvc\Reply\AbstractReply {
		$request = $controller->request();
		$session = $controller->session();

		$sessionKey = 'tfaData_dbtech_security_webauthn';

		if ($request->isPost())
		{
			if ($request->filter('setup', 'bool'))
			{
				$newProviderData = $session->get($sessionKey);
				if (!is_array($newProviderData))
				{
					return null;
				}

				$currentConfig = $provider->getUserProviderConfig();
				$currentConfig = is_array($currentConfig) ? $currentConfig : [];

				$newProviderData = array_merge($newProviderData, $currentConfig);
				if (!$this->verify('setup', $user, $newProviderData, $request))
				{
					return $controller->error(\XF::phrase('two_step_verification_value_could_not_be_confirmed'));
				}

				/** @var \XF\Repository\Tfa $tfaRepo */
				$tfaRepo = Helper::repository(\XF\Repository\Tfa::class);
				$tfaRepo->updateUserTfaData($user, $provider, $newProviderData);

				$session->remove($sessionKey);
			}
			elseif ($request->filter('remove', 'bool'))
			{
				$currentConfig = $provider->getUserProviderConfig();
				$currentConfig = is_array($currentConfig) ? $currentConfig : [];

				if (!empty($currentConfig['credentialStorage']))
				{
					$keysToRemove = $request->filter('keys', 'array-str');
					foreach ($keysToRemove as $key)
					{
						unset($currentConfig['credentialStorage'][$key]);
					}

					if (empty($currentConfig['credentialStorage']))
					{
						/** @var \XF\Entity\UserTfa|null $userTfa */
						$userTfa = $provider->UserEntries[\XF::visitor()->user_id];
						if ($userTfa)
						{
							$userTfa->delete();
						}
					}
					else
					{
						/** @var \XF\Repository\Tfa $tfaRepo */
						$tfaRepo = Helper::repository(\XF\Repository\Tfa::class);
						$tfaRepo->updateUserTfaData($user, $provider, $currentConfig);
					}
				}

				$session->remove($sessionKey);
			}

			return null;
		}
		else
		{
			$newProviderData = $this->generateInitialData($user);
			$newTriggerData = $this->trigger('setup', $user, $newProviderData, $request);

			$session->set($sessionKey, $newProviderData);
		}

		$viewParams = [
			'provider' => $provider,
			'user' => $user,
			'providerData' => $config,
			'newProviderData' => $newProviderData,
			'newTriggerData' => $newTriggerData
		];
		return $controller->view(
			\DBTech\Security\Pub\View\Account\TwoStepWebAuthnManage::class,
			'dbtech_security_account_two_step_webauthn_manage',
			$viewParams
		);
	}

	/**
	 * @param \Webauthn\PublicKeyCredentialSourceRepository $repo
	 *
	 * @return Server
	 */
	protected function getCredentialsServer(\Webauthn\PublicKeyCredentialSourceRepository $repo): Server
	{
		$options = \XF::options();
		$provider = $this->getProvider();
		$metadataRepo = null;
		if (!empty($provider->options['key_pin']) && $this->enableMds)
		{
			$metadataRepo = new SimpleMetadataStatementRepository(new FilesystemAdapter(
				'dbtechSecurity',
				0,
				\XF::app()
					->config('internalDataPath')
			));
		}

		// RP Entity
		$rpEntity = new PublicKeyCredentialRpEntity(
			!empty($provider->options['site_name']) ? $provider->options['site_name'] : $options->boardTitle,
			parse_url($options->boardUrl, PHP_URL_HOST),
			!empty($provider->options['site_icon']) ? $provider->options['site_icon'] : null
		);

		return new Server(
			$rpEntity,
			$repo,
			$metadataRepo
		);
	}

	/**
	 * @param \XF\Entity\User $user
	 *
	 * @return PublicKeyCredentialUserEntity
	 */
	protected function getUserEntity(\XF\Entity\User $user): PublicKeyCredentialUserEntity
	{
		// User Entity
		return new PublicKeyCredentialUserEntity(
			$user->username,
			$user->user_id,
			$user->username,
			\XF::canonicalizeUrl($user->getAvatarUrl(null))
		);
	}
}