<?php

namespace DBTech\Security\Watcher;

use DBTech\Security\Entity\Watcher;
use DBTech\Security\Helper;
use XF\Entity\User;
use XF\Mvc\Entity\ArrayCollection;

/**
 * Class AbstractHandler
 *
 * @package DBTech\Security\Watcher
 */
abstract class AbstractHandler
{
	protected array $options = [
		'closeForum' => true,
		'emailWebmaster' => true,
		'banIp' => true,
		'banUser' => true,
		'emailUser' => false,
		'lockChange' => false,
		'lockReset' => false,
		'lockUser' => false,
		'adminLockUser' => false
	];
	protected string $contentType;
	protected \XF\Language $defaultLanguage;

	/** @var \DBTech\Security\Entity\Watcher[]|ArrayCollection */
	protected $watchers;


	/**
	 * @param \DBTech\Security\Entity\Watcher $watcher
	 *
	 * @return \XF\Phrase
	 */
	abstract public function getParsedRule(Watcher $watcher): \XF\Phrase;

	/**
	 * @param \DBTech\Security\Entity\Watcher $watcher
	 * @param array $params
	 * @param User|null $user
	 * @param string|null $logMessage
	 *
	 * @return bool
	 */
	abstract protected function checkWatcher(
		Watcher $watcher,
		array &$params = [],
		?User $user = null,
		?string &$logMessage = null
	): bool;

	/**
	 * AbstractHandler constructor.
	 *
	 * @param string $contentType
	 */
	public function __construct(string $contentType)
	{
		$this->contentType = $contentType;

		$this->init();
	}

	/**
	 * Designed to be overridden if need be
	 */
	protected function init()
	{
		$watcherRepo = Helper::repository(\DBTech\Security\Repository\Watcher::class);

		$allWatchers = $watcherRepo->getWatchers();
		if ($allWatchers === null)
		{
			$watcherRepo->rebuildCache();
		}
		elseif (!empty($allWatchers[$this->getContentType()]))
		{
			$this->watchers = $allWatchers[$this->getContentType()];
		}
		else
		{
			$this->watchers = $this->em()->getEmptyCollection();
		}

		$this->defaultLanguage = $this->app()->language($this->options()->defaultLanguageId);
	}

	/**
	 * @return array
	 */
	public function getOptions(): array
	{
		return $this->options;
	}

	/**
	 * @param string $key
	 *
	 * @return mixed|null
	 */
	public function getOption(string $key): ?bool
	{
		return $this->options[$key] ?? null;
	}

	/**
	 * @return string
	 */
	public function getContentType(): string
	{
		return $this->contentType;
	}

	/**
	 * @param ArrayCollection|null $watchers
	 */
	public function setWatchers(?ArrayCollection $watchers = null)
	{
		$this->watchers = $watchers;
	}

	/**
	 * @param bool $onlyViewable
	 * @param bool $force
	 *
	 * @return Watcher[]|ArrayCollection
	 */
	public function getWatchers(bool $onlyViewable = false, bool $force = false)
	{
		if ($this->watchers === null || $force)
		{
			/** @var \DBTech\Security\Entity\Watcher[]|ArrayCollection $watchers */
			$watchers = Helper::finder(\DBTech\Security\Finder\Watcher::class)
				->where('watcher_type', $this->getContentType())
				->fetch()
			;

			$this->watchers = $watchers;
		}

		if ($onlyViewable)
		{
			return $this->watchers->filterViewable();
		}

		return $this->watchers;
	}

	/**
	 * @return \XF\Phrase
	 */
	public function getTitle(): \XF\Phrase
	{
		return \XF::phrase('dbtech_security_watcher_title.' . $this->contentType);
	}

	/**
	 * @return \XF\Phrase
	 */
	public function getDescription(): \XF\Phrase
	{
		return \XF::phrase('dbtech_security_watcher_description.' . $this->contentType);
	}

	/**
	 * @return bool
	 */
	public function isActive(): bool
	{
		return true;
	}

	/**
	 * @param array $params
	 * @param User|null $user
	 *
	 * @throws \XF\PrintableException
	 */
	public function trigger(array $params = [], ?User $user = null)
	{
		if (!$this->preCheck($params, $user))
		{
			return;
		}

		$watchers = $this->getWatchers(true);
		if (!$watchers->count())
		{
			return;
		}

		foreach ($watchers as $watcher)
		{
			$this->triggerWatcher($watcher, $params, $user);
		}
	}

	/**
	 * @param array $params
	 * @param User|null $user
	 *
	 * @return bool
	 */
	protected function preCheck(array &$params, ?User $user = null): bool
	{
		return true;
	}

	/**
	 * @param \DBTech\Security\Entity\Watcher $watcher
	 * @param array $params
	 * @param User|null $user
	 *
	 * @throws \XF\PrintableException
	 */
	public function triggerWatcher(Watcher $watcher, array $params = [], ?User $user = null)
	{
		$logMessage = null;
		if ($this->checkWatcher($watcher, $params, $user, $logMessage))
		{
			$watcherService = Helper::service(\DBTech\Security\Service\Watcher\Trigger::class, $watcher, $user);
			if ($logMessage)
			{
				$watcherService->setLogMessage($logMessage);
			}
			$watcherService->setParams($params);
			if ($watcherService->trigger())
			{
				$this->postTrigger($watcher, $params, $user, $logMessage);
			}
		}
	}

	/**
	 * @param \DBTech\Security\Entity\Watcher $watcher
	 * @param array $params
	 * @param User|null $user
	 * @param string|null $logMessage
	 */
	protected function postTrigger(Watcher $watcher, array $params = [], ?User $user = null, ?string $logMessage = null)
	{
	}

	/**
	 * @param \DBTech\Security\Entity\Watcher $watcher
	 *
	 * @return string
	 */
	public function renderOptions(Watcher $watcher): string
	{
		$templateName = $this->getOptionsTemplate();
		if (!$templateName)
		{
			return '';
		}
		return $this->app()->templater()->renderTemplate(
			$templateName,
			array_merge($this->getDefaultTemplateParams('options'), ['watcher' => $watcher])
		);
	}

	/**
	 * @return string|null
	 */
	public function getOptionsTemplate(): ?string
	{
		return 'admin:dbtech_security_watcher_edit_' . $this->contentType;
	}

	/**
	 * @param array $input
	 *
	 * @return array
	 */
	public function filterOptions(array $input = []): array
	{
		return $input;
	}

	/**
	 * @param string $context
	 *
	 * @return array
	 */
	protected function getDefaultTemplateParams(string $context): array
	{
		return [
			'title' => $this->getTitle(),
			'options' => $this->options
		];
	}

	/**
	 * @param string $type
	 * @param bool $throw
	 *
	 * @return AbstractHandler|null
	 * @throws \Exception
	 */
	protected function getHandler(string $type, bool $throw = true): ?AbstractHandler
	{
		return Helper::repository(\DBTech\Security\Repository\Watcher::class)
			->getHandler($type, $throw)
		;
	}

	/**
	 * @return \ArrayObject
	 */
	protected function options(): \ArrayObject
	{
		return \XF::app()->options();
	}

	/**
	 * @return \XF\Mvc\Entity\Manager
	 */
	protected function em(): \XF\Mvc\Entity\Manager
	{
		return \XF::app()->em();
	}

	/**
	 * @return \XF\App
	 */
	protected function app(): \XF\App
	{
		return \XF::app();
	}
}