<?php

namespace DBTech\Security\Watcher;

use DBTech\Security\Entity\Watcher;
use DBTech\Security\Helper;
use XF\Entity\User;

/**
 * Class AdminPermissions
 *
 * @package DBTech\Security\Watcher
 */
class AdminPermissions extends AbstractHandler
{
	/**
	 * @param Watcher $watcher
	 *
	 * @return \XF\Phrase
	 */
	public function getParsedRule(Watcher $watcher): \XF\Phrase
	{
		$language = \XF::language();

		return \XF::phrase('dbtech_security_x_changes_same_user_y_hours', [
			'changes' => $language->numberFormat($watcher->rule_data['intrusions']),
			'hours' => $language->numberFormat($watcher->rule_data['hours'])
		]);
	}

	/**
	 * @param array $params
	 * @param User|null $user
	 *
	 * @return bool
	 */
	protected function preCheck(array &$params, ?User $user = null): bool
	{
		if (empty($params['ipaddress'])
			|| !isset($params['script'])
			|| !isset($params['id'])
			|| !isset($params['field'])
			|| !isset($params['differences'])
		) {
			return false;
		}

		/** @var \XF\Repository\ChangeLog $changeRepo */
		$changeRepo = Helper::repository(\XF\Repository\ChangeLog::class);
		$changeRepo->logChange(
			'dbtech_security_adminperm',
			$params['id'],
			'permissions',
			$params['old'],
			$params['new'],
			\XF::visitor()->user_id
		);

		return true;
	}

	/**
	 * @param Watcher $watcher
	 * @param array $params
	 * @param User|null $user
	 * @param string|null $logMessage
	 *
	 * @return bool
	 */
	protected function checkWatcher(
		Watcher $watcher,
		array &$params = [],
		?User $user = null,
		?string &$logMessage = null
	): bool {
		if (empty($watcher->rule_data['intrusions']
			|| empty($watcher->rule_data['hours']))
		) {
			return false;
		}

		$changeLogFinder = Helper::finder('XF:ChangeLog')
			->where('content_type', 'dbtech_security_adminperm')
			->where('edit_user_id', \XF::visitor()->user_id)
			->where('edit_date', '>=', \XF::$time - (3600 * $watcher->rule_data['hours']))
		;
		if ($changeLogFinder->total() != $watcher->rule_data['intrusions'])
		{
			return false;
		}

		$logMessage = '<li>' . $this->defaultLanguage->renderPhrase('dbtech_security_' . $params['script'], [
				'param1' => $this->defaultLanguage->dateTime(\XF::$time),
				'param2' => $params['field'],
				'param3' => $params['differences'],
				'param4' => $params['ipaddress']
			]) . '</li>';

		return true;
	}

	/**
	 * @param array $input
	 *
	 * @return array
	 */
	public function filterOptions(array $input = []): array
	{
		return $this->app()->inputFilterer()->filterArray($input, [
			'intrusions' => 'uint',
			'hours' => 'uint',
		]);
	}
}